# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 14:56:13 2015
The Huber_braun neuronal model function
@author: porio
"""

import numpy as np
from numba import jit,float64

# PARAMETERS -- Must be defined first to allow numba compile
# Node parameters
# Any of them can be redefined as a vector of length nnodes
gd = 2.5; gr = 2.8; gsd = 0.21; gsr = 0.28;
gl = 0.06; gh = 0.4;
V0d = -25.; V0r = -25.; zd = 0.25; zr = 0.25;tr = 2.;
V0sd = -40.; zsd = 0.11; tsd = 10.;
eta = 0.014; kappa = 0.18; tsr = 35.;
V0h= -85.; zh = -0.14; th=125.;
Ed = 50.; Er = -90.; El = -80.; Eh = -30.;
temp=36.

rho = 1.3**((temp-25.)/10)
phi = 3**((temp-25.)/10)

Area=15000
gusd = gusr = 20. #pS
guh = 2.  #pS

Nsd = 10*gsd*Area/gusd
Nsr = 10*gsr*Area/gusr
Nh = 10*gh*Area/guh

# Network parameters
nnodes=50
Ggj=0.01

# Mock connectivity only for the purpose of compilation
CM=np.random.binomial(1,0.1,(nnodes,nnodes))

def InitVars(v=None):
    """
    Calculate initial values of variables, starting from a initial voltage.
    
    Initial voltage can be given as a single value (equal for all nodes), one
    value per node or random.

    Parameters
    ----------
    v : float, 1D ndarray or None, optional.
    
        If float, the same initial values are calculated per each node.
        If array or list (length must be nnodes), a different set of initial
        condition is calculated per node.         
        If None, a random voltage in [-65,60] is calculated for each node.
        The default is None.

    Returns
    -------
    y0 : ndarray, 5 x nnodes
    
        Contains the initial values for the _nnodes_ nodes. The order is
        voltage, asd, asr, ah, ar.

    """
    global rho,phi
    if v is None:
        v=np.random.uniform(low=-65,high=-60,size=nnodes)
    
    #Luego calculamos el valor de las variables a ese voltaje
    ar = 1/(1+np.exp(-zr*(v-V0r)));
    asd = 1/(1+np.exp(-zsd*(v-V0sd)));
    ah= 1/(1+np.exp(-zh*(v-V0h)))
    
    rho = 1.3**((temp-25.)/10)
    phi = 3**((temp-25.)/10)
    asr = -eta*rho*gsd*asd*(v - Ed)/kappa;
    
    #initial values Vector
    y0=np.array([v,asd,asr,ah,ar])
    
    return y0

y=InitVars()
runInt=0.025  # intervalo para resolver la SDE

@jit(float64[:,:](float64[:,:],float64),nopython=True)
def HyB(y,t):
    """
    Calculate the derivatives (deterministic) of the HBIh model at time t.

    Parameters
    ----------
    y : ndarray, 5 x nnodes
        Variables at time t.
    t : float
        Time.

    Returns
    -------
    Det : ndarray, 5 x nnodes
        Derivatives.

    """
    v=y[0,:]
    asd=y[1,:]
    asr=y[2,:]
    ah=y[3,:]
    ar=y[4,:]
    
    ad = 1/(1+np.exp(-zd*(v-V0d)))
    isd = rho*gsd*asd*(v - Ed)
    psr = (asr**2)/(asr**2+0.4**2)
    Imemb=isd + rho*gd*ad*(v - Ed) + rho*(gr*ar + gsr*psr)*(v-Er) \
                + rho*gh*ah*(v - Eh)+ rho*gl*(v - El)
    arinf = 1/(1+np.exp(-zr*(v-V0r)))
    asdinf = 1/(1+np.exp(-zsd*(v-V0sd)))
    ahinf= 1/(1+np.exp(-zh*(v-V0h)))

    Igj = np.sum(CM * Ggj * (np.expand_dims(v,1) - v),-1)

    Det=np.vstack((-Imemb - Igj,
                phi*(asdinf - asd)/tsd,
                phi*(-eta*isd - kappa*asr)/tsr,
                phi*(ahinf-ah)/th,
                phi*(arinf - ar)/tr))
    
    Stoch=np.vstack((rho*gsr*(v-Er)*np.sqrt(psr*(1-psr)/Nsr),
                    np.sqrt((np.abs(asdinf*(1-asd)+asd*(1-asdinf))*phi)/(tsd*Nsd)),
                    np.zeros(nnodes),
                    np.sqrt((np.abs(ahinf*(1-ah)+ah*(1-ahinf))*phi)/(th*Nh)),
                    np.zeros(nnodes))) * np.random.randn(5,nnodes) / np.sqrt(runInt)
    
    Det += Stoch
    
    return Det

@jit(float64[:,:](float64[:,:],float64),nopython=True)
def HyBdet(y,t):
    """
    Calculate the derivatives (stochastic) of the HBIh model at time t.
    
    Noise intensity is given by the module-wide variables Nsd, Nsr, Nh

    Parameters
    ----------
    y : ndarray, 5 x nnodes
        Variables at time t.
    t : float
        Time.

    Returns
    -------
    Det : ndarray, 5 x nnodes
        Derivatives.

    """
    v=y[0,:]
    asd=y[1,:]
    asr=y[2,:]
    ah=y[3,:]
    ar=y[4,:]

    ad = 1/(1+np.exp(-zd*(v-V0d)))
    isd = rho*gsd*asd*(v - Ed)
    psr = (asr**2)/(asr**2+0.4**2)
    Imemb=isd + rho*gd*ad*(v - Ed) + rho*(gr*ar + gsr*psr)*(v-Er) \
                + rho*gh*ah*(v - Eh)+ rho*gl*(v - El)
    arinf = 1/(1+np.exp(-zr*(v-V0r)))
    asdinf = 1/(1+np.exp(-zsd*(v-V0sd)))
    ahinf= 1/(1+np.exp(-zh*(v-V0h)))

    Igj = np.sum(CM * Ggj * (np.expand_dims(v,1) - v),-1)

    Det=np.vstack((-Imemb - Igj,
                phi*(asdinf - asd)/tsd,
                phi*(-eta*isd - kappa*asr)/tsr,
                phi*(ahinf-ah)/th,
                phi*(arinf - ar)/tr))
    
    return Det

"""
The main function is starting from here          
"""



#Simulation parameters
adaptTime=30000  #length (in ms) of adaptation simulation for transient removal
adaptInt=0.05    #time interval (in ms) of adaptation simulation

runTime=202000  # length (in ms) of simulation 
runInt=0.025  # interval (in ms) for solving the ODE/SDE
sampling=0.25  # interval (in ms) for data storage and return


def SimAdapt():
    """
    Rus a deterministic simulation of length HB.adaptTime and dt=HB.adaptInt.
    
    Initial conditions are given by the the function HB.Initvars().
    """
    # calculate initial conditions. We are not giving any argument, thus the
    # ICs will be random
    y=InitVars()
    
    Tadapt=np.arange(0,adaptTime,adaptInt)
    print("Adaptation simulation...  %g ms %d steps"%(adaptTime,len(Tadapt)))
    for t in Tadapt:
        y+=adaptInt*HyBdet(y,t)
    return y

def Sim(y0=None,Stoch=False,verbose=False):
    """
    Simulate the model with the defined parameters.

    Parameters
    ----------
    y0 : float or ndarray of shape (nnodes, ) or (5,nnodes) , optional
        Initial conditions. If float, this value is taken as the initial 
        voltage for all nodes and the rest of variables is calculated as
        the equilibrium at that voltage. If ndarray of single dimension
        (must be length nnodes), each value is given for each node and the 
        rest of variables is calculated at equilibrium. If ndarray of shape
        (5, nnodes), it specifies the whole set of initial conditions.
        If None, random initial voltages in [-65,-60] are drawn for each node.
        The default is None.
    Stoch : Boolean, optional
        If True, the simulation is stochastic. The default is False.
    verbose : Boolean, optional
        If True, messages are printed every 1000 ms of simulation.
        The default is False.

    Raises
    ------
    ValueError
        An initial check is performed to check the size of CM and nnodes.
        If they don't match a ValueError is raised

    Returns
    -------
    Y_t : ndarray
        The calculated variable trajectories. Variables (v, asd, asr, ah, ar) 
        go in the first dimension and time in the last dimension
    T_run
        Vector containing the time values of the trajectories.

    """
    if CM.shape[0]!=CM.shape[1] or CM.shape[0]!=nnodes:
        raise ValueError("check CM dimensions (",CM.shape,") and number of nodes (",nnodes,")")
    
    global Nsd,Nsr,Nh
    
    Nsd = 10*gsd*Area/gusd
    Nsr = 10*gsr*Area/gusr
    Nh = 10*gh*Area/guh
    
    # ********************************************************************
    # *** The following 2 lines are REALLY important before simulation ***
    # *** No parameter can be changed after                            ***
    # ********************************************************************
    HyB.recompile()
    HyBdet.recompile()

    if type(y0)==np.ndarray:
        if len(y0.shape)==1:
            y=InitVars(y0)
        else:
            y=y0
    elif y0 is None:
        y=SimAdapt()
    else:
        y=InitVars(y0*np.ones(nnodes))
         
    s_ratio=int(sampling/runInt)  # sampling ratio
    Trun=np.arange(0,runTime,runInt)
    Y_t=np.zeros((len(Trun)//s_ratio,nnodes))  #Vector para guardar datos

    print("Simulating %g ms dt=%g, Total %d steps"%(runTime,runInt,len(Trun)))

    if verbose and not Stoch:
        for i,t in enumerate(Trun):
            if i%s_ratio==0:
                Y_t[i//s_ratio]=y[0]
            if t%1000==0:
                print("%g of %g ms"%(t,runTime))
            y += runInt*HyBdet(y,t)

    if verbose and Stoch:
        for i,t in enumerate(Trun):
            if i%s_ratio==0:
                Y_t[i//s_ratio]=y[0]
            if t%1000==0:
                print("%g of %g ms"%(t,runTime))
            y += runInt*HyB(y,t)            

    if not verbose and not Stoch:
        for i,t in enumerate(Trun):
            if i%s_ratio==0:
                Y_t[i//s_ratio]=y[0]
            y += runInt*HyBdet(y,t)

    if not verbose and Stoch:
        for i,t in enumerate(Trun):
            if i%s_ratio==0:
                Y_t[i//s_ratio]=y[0]
            y += runInt*HyB(y,t)           
            
    return Y_t,Trun[::s_ratio]
        
def ParamsNode():
    """
    Return current node parameters.

    Returns
    -------
    pardict : Dictionary

    """
    pardict={}
    for var in ('gd','gr','gsd','gsr','gl','gh','V0d','V0r','zd','zr',
                'tr','V0sd','zsd','tsd','eta','kappa','tsr','V0h','zh',
                'th','Ed','Er','El','Eh','temp','Area','gusd','gusr','guh'):
        pardict[var]=eval(var)
        
    return pardict

def ParamsNet():
    """
    Return current network parameters.

    Returns
    -------
    pardict : Dictionary
        Number of nodes, Coupling constant, Connectivity matrix.

    """
    pardict={}
    for var in ('nnodes','Ggj','CM'):
        pardict[var]=eval(var)
        
    return pardict

def ParamsSim():
    """
    Return simulation parameters.

    Returns
    -------
    pardict : Dictionary
        Adaptation and simulation times and intervals.

    """
    pardict={}
    for var in ('adaptTime','adaptInt','runTime','runInt','sampling'):
        pardict[var]=eval(var)
        
    return pardict

#%%
if __name__=="__main__":
    
    import matplotlib.pyplot as plt
    
    adaptTime=1000
    runTime=2000
    Ggj=0.02
    
    nnodes=30
    CM=np.random.binomial(1,0.3,(nnodes,nnodes))
    gsd=np.random.uniform(0.18,0.23,size=nnodes)
    gh=np.random.uniform(0.3,0.6,size=nnodes)
    
    Vtrace,time=Sim(verbose=True,Stoch=True)
    
    plt.figure(1)
    plt.clf()
    plt.plot(time,Vtrace)
    
    



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 14:56:13 2015
Wilson Cowan model with noise and plasticity
@author: porio
"""

from matplotlib.pyplot import getp
from numba.core.types.misc import Module
import numpy as np
from numba import jit,float64,boolean, vectorize

from numba.core.errors import NumbaPerformanceWarning
import warnings
warnings.simplefilter('ignore', category=NumbaPerformanceWarning)

#Node parameters
# Any of them can be redefined as a vector of length nnodes
#excitatory connections
a_ee=3.5; a_ie_0=2.5
#inhibitory connections
a_ei=3.75; a_ii=0
#tau
tauE=0.010; tauI=0.020  # Units: seconds
#external input
P = 0.4 # 0.4
Q = 0
# inhibitory plasticity
rhoE=0.14 # target mean value for E
tau_ip=2  #time constant for plasticity

rE,rI=0.5,0.5
mu = 1;sigma=0.25

D=0
# This is a mock value, it will get overwritten
sqdtD=0.1

#network parameters
N=50
G=0.001
CM=np.random.binomial(1,0.1,(N,N)).astype(np.float64)


@vectorize([float64(float64)],nopython=True)
def S(x):
    return (1/(1+np.exp(-(x-mu)/sigma)))

@jit(float64[:,:](float64,float64[:,:],boolean,boolean),nopython=True)
def wilsonCowan(t,X, noisy, plasticity):
    
    E,I,a_ie = X
    noise=(noisy)*np.random.normal(0,sqdtD,size=N)
    
    exc = (-E + (1-rE*E)*S(a_ee*E - a_ie*I + G*np.dot(CM,E) + P + noise))/tauE
    inh = (-I + (1-rI*I)*S(a_ei*E - a_ii*I ))/tauI
    plast = (plasticity)*(I*(E-rhoE)/tau_ip)
    
    return np.vstack((exc,inh,plast))
    
   
"""
The main function is starting from here          
"""

E0=0.1
I0=0.1

### Time units are seconds  ###
tTrans=100
tstop=100
dt=0.001    #interval for points storage
dtSim=0.0001   #interval for simulation (ODE integration)
downsamp=int(dt/dtSim)



def SimAdapt(Init=None, plasticity=False):
    """
    Runs a deterministc simulation of duration 'timeTrans'
    Runs two deterministic simulations of timeTrans. 
    
    First with tau_ip=0.05
    Second with tau_ip=tau_ip/2
    """

    global tau_ip, timeTrans
    if Init is None:
        #print("initializing vars")
        Var=np.array([E0,I0,a_ie_0])[:,None]*np.ones((1,N))
    else:
        Var=Init
    # generate the vector again in case variables have changed
    timeTrans=np.arange(0,tTrans,dtSim)
    if plasticity:
        old_tau_ip=tau_ip
        tau_ip=0.05
    noisy = False
    wilsonCowan.recompile()

    
    
    # Varinit=np.zeros((len(timeTrans),3,N))
    for i,t in enumerate(timeTrans):
        # Varinit[i]=Var 
        # at the begining,  Var is boolean for soome reason
        Var+=dtSim*wilsonCowan(t,Var,noisy, plasticity)
    
    if plasticity:
        tau_ip=old_tau_ip/2
        wilsonCowan.recompile()
        
        for i,t in enumerate(timeTrans):
            Var+=dtSim*wilsonCowan(t,Var,noisy,plasticity)
        tau_ip=old_tau_ip
    
    return Var

def Sim(Var0=None,verbose=False, noisy = False, plasticity = False):
    """
    Run a network simulation with the current parameter values.
    
    If D==0, run deterministic simulation.
    Note that the time unit in this model is seconds.

    Parameters
    ----------
    Var0 : ndarray (X,N), ndarray (X,), or None
        Initial values. Either one value per each node (X,N), 
        one value for all (X,) or None. If None, an equilibrium simulation
        is run with faster plasticity kinetics. The default is None.
        X is the dimension of the variables. If X==2, the model is without
        plasticity. If X==3, model is with plasticity
    verbose : Boolean, optional
        If True, some intermediate messages are shown.
        The default is False.
    noisy: boolean

    Raises
    ------
    ValueError
        An error raises if the dimensions of CM and the number of nodes
        do not match.

    Returns
    -------
    Y_t : ndarray
        Time trajectory for the three variables of each node.
    time : TYPE
        Values of time.

    """
    global CM,sqdtD,downsamp
    
    #Connectivity matrix assertion
    if CM.shape[0]!=CM.shape[1] or CM.shape[0]!=N:
        raise ValueError("check CM dimensions (",CM.shape,") and number of nodes (",N,")")
    
    if CM.dtype is not np.dtype('float64'):
        try:
            CM=CM.astype(np.float64)
        except:
            raise TypeError("CM must be of numeric type, preferred float")
    
    if plasticity:
        if (rhoE==None) or (tau_ip==None):
            raise ValueError("Plastic variables are not iniatilized!")
    
    #variable initialization
    if type(Var0)==np.ndarray:
        if len(Var0.shape)==1:
            Var=Var0*np.ones((1,N))
        else:
            Var=Var0
    elif Var0 is None:
        Var=SimAdapt(None,plasticity)

    timeSim=np.arange(0,tstop,dtSim)
    time=np.arange(0,tstop,dt)
    downsamp=int(dt/dtSim)
    
    Y_t=np.zeros((len(time),3,N))

    if verbose:
        print("Simulating %g s dt=%g, Total %d steps"%(tstop,dtSim,len(timeSim)))

    if noisy:
        sqdtD=D/np.sqrt(dtSim)
    wilsonCowan.recompile()
    
    for i,t in enumerate(timeSim):
        if i%downsamp==0:
            Y_t[i//downsamp]=Var
        Var += dtSim*wilsonCowan(t,Var, noisy, plasticity)
        if verbose:
            if t%10==0:
                print("%g of %g ms"%(t,tstop))
            
    return Y_t,time

def getParams(pars = 'node'):

    if(pars=='network'):
        params = ('N','G','CM')
    elif(pars=='simulation'):
        params = ('tTrans','tstop','dt','dtSim')
    elif(pars=='node'):
        params = ('a_ee','a_ei','a_ii','tauE','tauI',
                'P','Q','rhoE','tau_ip','rE','rI','mu','sigma')
    pardict={}
    for var in params:
        pardict[var]=eval(var)
    return pardict


#%%
if __name__=="__main__":
    
    import matplotlib.pyplot as plt
    
    tTrans=25
    tstop=10
    G=0.005
    D=0
    rhoE=0.14
    
    N=20
    CM=np.random.binomial(1,0.3,(N,N)).astype(np.float32)
    P=np.random.uniform(0.3,0.5,N)
    # P=.4
    
    Vtrace,time=Sim(verbose=True)
    getParams(pars='node')
#%%    
    plt.figure(1)
    plt.clf()
    plt.subplot(211)
    plt.plot(time,Vtrace[:,0,:])

    plt.subplot(212)
    plt.plot(time,Vtrace[:,2,:])
    
    



# -*- coding: utf-8 -*-
"""
Created on Mon Oct  8 21:37:23 2018

@author: patri
"""

import numpy as np
import matplotlib.pyplot as plt
from numba import jit,float64


Vr=-70
DT=2
Vth=-50
Vreset=-58
Vspk=0
gL= 6
C=200
tau_exc=5; Eexc=0 
tau_inh=20; Einh=-70

# Main function
@jit(float64[:,:](float64[:,:],float64[:]),nopython=True)
def LIF(X,I):
    """
    Network of Exponential-Integrate-and-Fire
    
    Calculates the derivatives of voltage and synaptic conductances at a given time.

    Parameters
    ----------
    X : array of float64 of size (3,N)
        N is the number of neurons. Accross the first dimension, the variables are Voltage, excitatory conductance, inhibitory conductance.
    I : array of float64, 1-D and length N 
        External current applied to each neuron.

    Returns
    -------
    array of float64 of size (3,N)
        The derivative at time _t_ for each variable.

    """
    v,ge,gi=X
    return np.vstack(((-gL*(v-Vr)+DT*np.exp((v-Vth)/DT)-ge*(v-Eexc)-gi*(v-Einh)+I)/C,
                     -ge / tau_exc,
                     -gi / tau_inh))

#names of parameters
    
N_i=[16,10] #Numero de neuronas [excitatorias,inhibitorias, ...]
N=sum(N_i) #numero de neuronas a simular

#individual parameter values
Vr_i=[-65,-62] #-70
DT_i=[0.8, 3] #2
Vth_i=[-52,-42] #-50
Vreset_i=[-53, -54] #-58
Vspk_i=0
gL_i=[4.3, 2.9]# 6
C_i=[104, 30] #200

Params=["Vr","DT","Vth","Vreset","Vspk","gL","C"]
"""
We will take the parameters as lists of lenght K (number of neuron
types) and convert them to arrays of length N (number of neurons).
"""    
def InitPars():
    """
    
    Creates arrays of parameters of length N (number of neurons, one parameter per neuron)
    from arrays of length K (number of neuron types). If any of the given values is a scalar, 
    it will remain a scalar. **This function does not take any argument**;
    it operates from the module-wide existing variables 
    
    Vr_i    Membrane (leak) reversal potential
    
    DT_i    ΔT parameter of EIF neuron
    
    Vth_i    V_th parameter of EIF neuron
    
    Vreset_i     Voltage value after a spike (reset voltage)
    
    Vspk_i     Voltage for detecting spikes
    
    gL_i    Membrane (leak) conductance
    
    C_i     Membrane capacitance

    These variables must exist either as a scalar or as a list of length K (number of neuron types)

    Raises
    ------
    ValueError
        If any of the variables is of a different length than the neuron types.

    Returns
    -------
    None. The arrays of variables become available module-wide.

    """
    global Vr,DT,Vth,Vreset,Vspk,gL,C
    
    for param in Params:
        if type(eval(param+"_i"))==list:
            if eval("len("+param+"_i)")==len(N_i):
                exec("%s = np.array([%s_i[0],]*N_i[0] + [%s_i[1],]*N_i[1])"%((param,)*3), globals())
            else:
                raise ValueError("check the length of variable"+param)
        else:
            exec("%s = %s_i"%((param,)*2))
    
#NETWORK parameters    
#Synaptic Parameters
tau_exc=5; Eexc=0 
tau_inh=20; Einh=-70

#CONNECTIVITY MATRIX
# By default, all excitatory and inhibitory weigths will be the same. This
# doesn't have to be the case; just use a heterogeneous Connectivity matrix.
WsynE = 0.2; WsynI = 0.3 #defaults synaptic weigths
#recordar que será siempre positivo porque son sinapsis de tipo conductancia

## DEFAULT CONNECTIVITY: Random connectivity matrices
def InitCM(Pee=0.4, Pei=0.6, Pie=0.8, Pii=0.4):
    """
    Calculate a random adjacency Matrix and store it in the module-wide variables CMe and CMi.
    
    CMe : Array of size (N,Ne). All excitatory inputs
    
    CMi : Array of size (N,Ni). All inhibitory inputs
    
    The (i,j) entry of CMe (CMi) matrix denotes the connection from the j-th excitatory
    (inhibitory) neuron to the i-th neuron

    Parameters
    ----------
    Pee : float, 0 <= Pee <= 1, optional
        Connection probability from excitatory to excitatory neurons. 
    Pei : float, 0 <= Pei <= 1, optional
        Connection probability from inhibitory to excitatory neurons. 
    Pie : float, 0 <= Pie <= 1, optional
        Connection probability from excitatory to inhibitory neurons. 
    Pii : float, 0 <= Pii <= 1, optional
        Connection probability from inhibitory to inhibitory neurons. 

    Returns
    -------
    None. CMe and CMi matrices are internally defined

    """
    global CMe,CMi,Ne,Ni
    Ne=N_i[0];Ni=N_i[1]
    CMee=WsynE * np.random.binomial(1,Pee,size=(Ne,Ne)) #from exc to exc
    CMie=WsynE * np.random.binomial(1,Pie,size=(Ni,Ne)) #from exc to inh
    CMei=WsynI * np.random.binomial(1,Pei,size=(Ne,Ni)) #from inh to exc
    CMii=WsynI * np.random.binomial(1,Pii,size=(Ni,Ni)) #from inh to exc
    CMe=np.r_[CMee,CMie]  #All inputs FROM exc
    CMi=np.r_[CMei,CMii]  #All inputs FROM inh

#simulation parameters
tstop=5000; dt=0.1

# Input (noisy) current
Ioffset=80 #media del estímulo
noise=2 #desviacion estandar del ruido

def InitVars(V0=-50):
    """
    Create a vector of initial variables. Initial voltage can be set, synaptic conductances
    will always be 0

    Parameters
    ----------
    V0 : float, optional
        Initial voltage. The default is -50.

    Returns
    -------
    X : array of floats, dimension (3,N)
        N is the number of neurons. 
        Accross the first dimension, the variables are Voltage, excitatory conductance, inhibitory conductance.

    """
    N=sum(N_i)
    v=V0*np.ones(N) #La variable v tiene dimensión nLIF
    ge=np.zeros_like(v)
    gi=np.zeros_like(v)
    X=np.array([v,ge,gi])
    return X

def runSim(X=None,recordV=False):
    """
    Run a simulation with the currently defined parameters.

    Parameters
    ----------
    X : 3xN Array of floats, optional
        Initial conditions. N is the number of neurons. Accross the first dimension,
        the variables are Voltage, excitatory conductance, inhibitory conductance.
        If None, ICs are set to V=-50, synaptic conductances=0. The default is None.
    recordV : boolean, optional
        If True, variables will be recorded for each neuron and returned at the end.
        Use with caution when simulating large networks. The default is False.

    Returns
    -------
    spikes : list of lists
        Each element of the list contains the index of the neuron and the time of firing.
    
    time : 1D array of floats
        Contains the times of the simulation
        
    X_t : T x 3 x N array of floats
        (Only returned if recordV==True) Time traces for voltage and synaptic conductances.
        T is the same length as time.

    """
    global N
    N=sum(N_i)
    if X is None:
        X=InitVars()
    time=np.arange(0,tstop,dt)
    spikes=[]
    # Istim=Ioffset+np.sqrt(noise/dt)*np.random.normal(size=(len(time),N)) #distinta corriente en cada neurona
    sqdt=np.sqrt(noise/dt)
    
    if recordV:
        X_t=np.zeros((len(time),3,N))
        for i,t in enumerate(time):
            Istim=Ioffset+sqdt*np.random.normal(size=N) #distinta corriente en cada neurona
            X_t[i,:]=X
            X+=dt*LIF(X,Istim)
            v=X[0]
            if any(v>=0):
                idx=np.where(v>=0)[0]
                X[0,idx] = Vreset[idx]
                for idx_i in idx:
                    spikes.append([idx_i,t])
                    if idx_i<Ne:
                        X[1] = X[1] + CMe[:,idx_i]
                    else:
                        X[2] = X[2] + CMi[:,idx_i-Ne]
        return spikes,time,X_t
    else:
        for i,t in enumerate(time):
            Istim=Ioffset+sqdt*np.random.normal(size=N) #distinta corriente en cada neurona
            X+=dt*LIF(X,Istim)
            v=X[0]
            if any(v>=0):
                idx=np.where(v>=0)[0]
                X[0,idx] = Vreset[idx]
                for idx_i in idx:
                    spikes.append([idx_i,t])
                    if idx_i<Ne:
                        X[1] = X[1] + CMe[:,idx_i]
                    else:
                        X[2] = X[2] + CMi[:,idx_i-Ne]
        return spikes,time


def ParamsNeuron():    
    pardict={}
    for var in ("Vr","DT","Vth","Vreset","Vspk","gL","C"):
        pardict[var]=eval(var)
        
    return pardict

def ParamsNeuronByType():
    pardict={}
    for var in ("Vr","DT","Vth","Vreset","Vspk","gL","C"):
        pardict[var]=eval(var+"_i")
        
    return pardict

def ParamsNet():
    pardict={}
    for var in ('N_i','N','tau_exc', 'Eexc', 'tau_inh', 'Einh', 'WsynE', 'WsynI', 'Ioffset', 'noise'):
        pardict[var]=eval(var)
    return pardict


def ParamsCMatrix():
    pardict={}
    for var in ('CMe','CMi'):
        pardict[var]=eval(var)
    return pardict

def ParamsSim():
    pardict={}
    for var in ('tstop','dt'):
        pardict[var]=eval(var)
        
    return pardict


#%%

if __name__=="__main__":

    N_i=[75,25]
    InitPars()
    InitCM()
    
    LIF.recompile()
    
    spikes,time,X_t=runSim(InitVars(),recordV=True)
      
    #%%
    
    
    if len(spikes)>0:
        spikes=np.array(spikes)
        neuron,fRates=np.unique(spikes[:,0],return_counts=True)
        fRates=fRates/(tstop/1000)            
                
    #%%
    # Vplot=np.arange(-65,-25,0.1)
    # plt.figure(2,figsize=(5,5))
    # plt.clf()
    # plt.plot(Vplot,F(Vplot))
    
    plt.figure(1,figsize=(12,6))
    plt.clf()
    if len(spikes)>0:
        plt.subplot(231)
        plt.plot(spikes[:,1],spikes[:,0],'.',ms=1)
        plt.xlim((0,tstop))
    plt.subplot(232)
    plt.plot(time[::10],X_t[::10,0,:Ne],'g')
    plt.plot(time[::10],X_t[::10,0,Ne:],'r')
    plt.subplot(233)
    plt.plot(time[::10],X_t[::10,1,:Ne],'g')
    plt.plot(time[::10],X_t[::10,1,Ne:],'r')
    plt.subplot(235)
    plt.plot(time[::10],X_t[::10,2,:Ne],'g')
    plt.plot(time[::10],X_t[::10,2,Ne:],'r')

    
    plt.subplot(234)
    maxV=np.max(np.abs(np.c_[CMe,CMi]))
    plt.imshow(np.c_[CMe*1,CMi*-1],cmap='bwr',vmin=-maxV,vmax=maxV)
    plt.colorbar()
    
    plt.tight_layout()



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 22 14:56:13 2015
The Huber_braun neuronal model function
@author: porio
"""

import numpy as np
from numba import jit,float64, vectorize

from numba.core.errors import NumbaPerformanceWarning
import warnings
warnings.simplefilter('ignore', category=NumbaPerformanceWarning)

#Node parameters
# Any of them can be redefined as a vector of length nnodes
#excitatory connections
a_ee=3.5; a_ie=2.5
#inhibitory connections
a_ei=3.75; a_ii=0
#tau
tauE=0.010; tauI=0.020  # Units: seconds
#external input
P = 0.4 # 0.4
Q = 0
# inhibitory plasticity

rE,rI=0.5,0.5
mu = 1;sigma=0.25

D=0
# This is a mock value, it will get overwritten
sqdtD=0.1

#network parameters
N=50
G=0.001
CM=np.random.binomial(1,0.1,(N,N)).astype(np.float64)


@vectorize([float64(float64)],nopython=True)
def S(x):
    return (1/(1+np.exp(-(x-mu)/sigma)))

@jit(float64[:,:](float64,float64[:,:]),nopython=True)
def wilsonCowan(t,X):
    E,I = X
    noise=np.random.normal(0,sqdtD,size=N)
    return np.vstack(((-E + (1-rE*E)*S(a_ee*E - a_ie*I + G*np.dot(CM,E) + P + noise))/tauE,
                     (-I + (1-rI*I)*S(a_ei*E - a_ii*I ))/tauI))

@jit(float64[:,:](float64,float64[:,:]),nopython=True)
def wilsonCowanDet(t,X):
    E,I = X
    return np.vstack(((-E + (1-rE*E)*S(a_ee*E - a_ie*I + G*np.dot(CM,E) + P))/tauE,
                     (-I + (1-rI*I)*S(a_ei*E - a_ii*I ))/tauI))

"""
The main function is starting from here          
"""

E0=0.1
I0=0.1

### Time units are seconds  ###
tTrans=2
tstop=10
dt=0.001    #interval for points storage
dtSim=0.0001   #interval for simulation (ODE integration)
downsamp=int(dt/dtSim)

# # adaptation time - note that adaptation occurs twice
# timeTrans=np.arange(0,tTrans,dtSim)
# # Simulation time
# timeSim=np.arange(0,tstop,dtSim)
# # time for storing data
# time=np.arange(0,dt,tstop)

def SimAdapt(Init=None):
    """
    Runs a deterministic simulation of timeTrans. 
    """
    global timeTrans
    if Init is None:
        Var=np.array([E0,I0])[:,None]*np.ones((1,N))
    else:
        Var=Init
    # generate the vector again in case variables have changed
    timeTrans=np.arange(0,tTrans,dtSim)

    wilsonCowanDet.recompile()
    
    # Varinit=np.zeros((len(timeTrans),3,N))
    for i,t in enumerate(timeTrans):
        # Varinit[i]=Var
        Var+=dtSim*wilsonCowanDet(t,Var)
    
    return Var

def Sim(Var0=None,verbose=False):
    """
    Run a network simulation with the current parameter values.
    
    If D==0, run deterministic simulation.
    Note that the time unit in this model is seconds.

    Parameters
    ----------
    Var0 : ndarray (2,N), ndarray (2,), or None
        Initial values. Either one value per each node (2,N), 
        one value for all (2,) or None. If None, an equilibrium simulation
        is run with faster plasticity kinetics. The default is None.
    verbose : Boolean, optional
        If True, some intermediate messages are shown.
        The default is False.

    Raises
    ------
    ValueError
        An error raises if the dimensions of CM and the number of nodes
        do not match.

    Returns
    -------
    Y_t : ndarray
        Time trajectory for the three variables of each node.
    time : TYPE
        Values of time.

    """
    global CM,sqdtD,downsamp
    
    if CM.shape[0]!=CM.shape[1] or CM.shape[0]!=N:
        raise ValueError("check CM dimensions (",CM.shape,") and number of nodes (",N,")")
    
    if CM.dtype is not np.dtype('float64'):
        try:
            CM=CM.astype(np.float64)
        except:
            raise TypeError("CM must be of numeric type, preferred float")
    
    
    if type(Var0)==np.ndarray:
        if len(Var0.shape)==1:
            Var=Var0*np.ones((1,N))
        else:
            Var=Var0
    elif Var0 is None:
        Var=SimAdapt()

    timeSim=np.arange(0,tstop,dtSim)
    time=np.arange(0,tstop,dt)
    downsamp=int(dt/dtSim)
         
    Y_t=np.zeros((len(time),2,N))  #Vector para guardar datos

    if verbose:
        print("Simulating %g s dt=%g, Total %d steps"%(tstop,dtSim,len(timeSim)))

    if verbose and D==0:
        wilsonCowanDet.recompile()
        for i,t in enumerate(timeSim):
            if i%downsamp==0:
                Y_t[i//downsamp]=Var
            if t%10==0:
                print("%g of %g s"%(t,tstop))
            Var += dtSim*wilsonCowanDet(t,Var)

    if verbose and D>0:
        sqdtD=D/np.sqrt(dtSim)
        wilsonCowan.recompile()
        for i,t in enumerate(timeSim):
            if i%downsamp==0:
                Y_t[i//downsamp]=Var
            if t%10==0:
                print("%g of %g ms"%(t,tstop))
            Var += dtSim*wilsonCowan(t,Var)

    if not verbose and D==0:
        wilsonCowanDet.recompile()
        for i,t in enumerate(timeSim):
            if i%downsamp==0:
                Y_t[i//downsamp]=Var
            Var += dtSim*wilsonCowanDet(t,Var)

    if not verbose and D>0:
        sqdtD=D/np.sqrt(dtSim)
        wilsonCowan.recompile()
        for i,t in enumerate(timeSim):
            if i%downsamp==0:
                Y_t[i//downsamp]=Var
            Var += dtSim*wilsonCowan(t,Var)
            
    return Y_t,time
        
def ParamsNode():
    pardict={}
    for var in ('a_ee','a_ei','a_ie','a_ii','tauE','tauI',
                'P','Q','rE','rI','mu','sigma'):
        pardict[var]=eval(var)
        
    return pardict

def ParamsNet():
    pardict={}
    for var in ('N','G','CM'):
        pardict[var]=eval(var)
        
    return pardict

def ParamsSim():
    pardict={}
    for var in ('tTrans','tstop','dt','dtSim'):
        pardict[var]=eval(var)
        
    return pardict
#%%
if __name__=="__main__":
    
    import matplotlib.pyplot as plt
    
    tTrans=25
    tstop=10
    G=0.005
    D=0
    
    N=20
    CM=np.random.binomial(1,0.3,(N,N)).astype(np.float32)
    P=np.random.uniform(0.34,0.5,N)
    # P=.4
    
    Vtrace,time=Sim(verbose=True)
    ParamsNode()
#%%    
    plt.figure(1)
    plt.clf()
    plt.subplot(211)
    plt.plot(time,Vtrace[:,0,:])

    plt.subplot(212)
    plt.plot(time,Vtrace[:,1,:])
    
    
